/* SPDX-License-Identifier: (GPL-2.0-only OR BSD-2-Clause) */

#ifndef _LINUX_BITOPS_H_
#define _LINUX_BITOPS_H_

#ifndef BITS_PER_LONG
# define BITS_PER_LONG __WORDSIZE
#endif

#define BITS_PER_BYTE 8

#endif
